import java.util.ArrayList;
import java.awt.Toolkit;

/******************************************************************************
*  Arena of battling gladiators (fighters).
*
*  @author   Daniel R. Collins (dcollins@superdan.net)
*  @since    2014-05-20
*  @version  1.09
******************************************************************************/

public class Arena {

   //--------------------------------------------------------------------------
   //  Constants
   //--------------------------------------------------------------------------

   public enum ArenaType {ManVsMan, ManVsMonster};
   final ArenaType ARENA_TYPE = ArenaType.ManVsMan;
	final XPTable.XPMode XP_MODE = XPTable.XPMode.Vol1;
   final int NUM_FIGHTERS = 1000;
   final int NUM_FIGHTS = 500;
   final int START_LEVEL = 0;
   final int MAX_LEVEL = 50;
   final int REPORT_LEVEL = 50;
   final int PCT_MAGIC_PER_LEVEL = 5;
   final int PCT_LEVEL_CHOICE_MOD = 100;
   final int MONSTER_MATRIX_MODIFIER = -4;
   final boolean ORIGINAL_XP = true;
   final boolean REPORT_ALL_FIGHTERS = false;
   final boolean REPORT_MONSTER_KILLS = false;
	final Armor.Type BASE_ARMOR = Armor.Type.Plate;

   //--------------------------------------------------------------------------
   //  Fields
   //--------------------------------------------------------------------------

	/** Arena type being run. */
	ArenaType arenaType;

	/** List of fighters for the arena. */
	Party fighterList;

	/** Desired size of the fighter list. */
	int desiredListSize;

	/** Base armor type for fighters. */
	Armor.Type baseArmorType;

	/** Fix primary weapon for fighters. */
	Weapon baseWeaponType;

   //--------------------------------------------------------------------------
   //  Constructors
   //--------------------------------------------------------------------------

   /**
   *  Constructor.
   */
   public Arena () {
		Dice.initialize();
		fighterList = new Party(); 
		XPTable.getInstance().setMode(XP_MODE);
		arenaType = ARENA_TYPE;
		desiredListSize = NUM_FIGHTERS;
		baseArmorType = BASE_ARMOR;
		baseWeaponType = null;
   }

   /**
   *  Constructor (set type, size).
   */
	public Arena (ArenaType type, int desiredSize) {
		this();
		arenaType = type;
		desiredListSize = desiredSize;
	}

   //--------------------------------------------------------------------------
   //  Methods
   //--------------------------------------------------------------------------

   /**
   *  Run the arena's top-level algorithm.
   */
   public void runArena () {

		// Run combats
      for (int cycle = 0; cycle < NUM_FIGHTS; cycle++) {
			runOneCycle();
      }
		
		// Report stuff
		reportFighterStatistics();
		reportAllFighters();
		reportMonsterKills();
   }

   /**
   *  Run one cycle of fights for the whole list.
   */
	public void runOneCycle () {
		recruitNewFighters();
		fighterList.shuffleMembers();
		fightDuels();
		fighterList.bringOutYourDead();
		fighterList.clearFallen();
		fighterList.healAll();
	}

   /**
   *  Fill out the fighter list.
   */
   void recruitNewFighters() {
      while (fighterList.size() < desiredListSize) {
         fighterList.add(newFighter(START_LEVEL));
      }
   }

   /**
   *  Create a new fighter of the indicated level.
   */
   Character newFighter (int level) {
		String name = NameGenerator.getInstance().getRandom();
      Character f = new Character(name, "Human", "Fighter", level, null); 
      f.setArmor(Armor.makeType(baseArmorType));
		Weapon primaryWeapon = (baseWeaponType == null) ?
			Weapon.randomPrimary() : new Weapon(baseWeaponType);
		f.drawWeapon(primaryWeapon);
		f.addEquipment(Weapon.randomSecondary());
		if (primaryWeapon.getHandsUsed() <= 1) {
	      f.setShield(Armor.makeType(Armor.Type.Shield));
		}
		if (arenaType == Arena.ArenaType.ManVsMonster) {
	      f.addEquipment(Weapon.silverDagger());
		}
      for (int i = 0; i <= level; i++) {
         f.checkMagicArmsBoost(PCT_MAGIC_PER_LEVEL);
      }
      return f;
   }

   /**
   *  Fight duels for all fighters in list.
   */
   void fightDuels () {
		switch (arenaType) {
			case ManVsMan: fightDuelsManVsMan(); return;
			case ManVsMonster: fightDuelsManVsMonster(); return;						
		}
   }

   /**
   *  Duel each pair of fighters.
   */
   void fightDuelsManVsMan () {
      for (int i = 0; i < fighterList.size() - 1; i += 2) {
			Monster m1 = fighterList.get(i);
			Monster m2 = fighterList.get(i+1);
			FightManager.fight(m1, m2);
			awardXP(m1, m2);
      }
   }

   /**
   *  Duel each fighter against a random monster.
   */
   void fightDuelsManVsMonster () {
      MonsterRandomizer ml = MonsterRandomizer.getInstance();
		for (int i = 0; i < fighterList.size(); i++) {
			Monster fighter = fighterList.get(i);
         int dungeonLevel = fighter.getLevel() 
            * PCT_LEVEL_CHOICE_MOD / 100;
			dungeonLevel = Math.max(1, dungeonLevel); 
         Monster monster = ml.randomMonsterOnDungeonLevel(
            dungeonLevel, MONSTER_MATRIX_MODIFIER);
			FightManager.fight(fighter, monster);
			awardXP(fighter, monster);
			if (!monster.horsDeCombat())
				addToKillTally(monster);
      }
   }

   /**
   *  Award XP for victory to winning monster.
   */
	void awardXP (Monster m1, Monster m2) {
		Monster winner = m1.horsDeCombat() ? m2 : m1;
		Monster loser = m1.horsDeCombat() ? m1 : m2;
		awardXP(winner, totalXPAward(loser));
	}

   /**
   *  Award XP and magic to one creature/character.
   */
	void awardXP (Monster m, int xp) {

		// Add XP
      int startLevel = m.getLevel();
      m.addXP(xp);
		
		// Check for level-up
		if (m.getLevel() > startLevel) {
         m.checkMagicArmsBoost(PCT_MAGIC_PER_LEVEL);
      }
	}

   /**
   *  Compute total XP award from monster + treasure.
   *  See DeltasDnD blog 2014-02-25, 2016-01-11.
   */
	int totalXPAward (Monster monster) {
		XPTable xpt = XPTable.getInstance();
		int monsterXP = monster.getXPAward();
   
      // Linear XP from Vol-1; treasure about 75%
      if (xpt.getMode() == XPTable.XPMode.Vol1) {
			return monsterXP * 4;
      }
      
      // Table XP from Sup-I; treasure from 95% to 70%.
      else if (xpt.getMode() == XPTable.XPMode.SupI) {
         double treasPercent = 0.95 - 0.01 * (monster.getHD() - 1);
         return (int) Math.round(1/(1 - treasPercent) * monsterXP);
      }
		
		// Unknown mode
		else {
			return 0;		
		}
   }

   /**
   *  Add to the kill tally for winning monster type.
   */
   void addToKillTally (Monster monster) {
		Monster prototype = MonsterList.getInstance()
			.getByType(monster.getType());
      if (prototype != null) {
      	prototype.addToKillTally(1);
      }
   }

   /**
   *  Print all of the fighters (for testing small groups).
   */
   void reportAllFighters() {
      if (REPORT_ALL_FIGHTERS) {
			fighterList.sortMembers();
			for (int i = 0; i < fighterList.size(); i++) {
				System.out.println(fighterList.get(i));
         }
         System.out.println();
      }
   }

   /**
   *  Generate and print statistics for all the fighters.
   */
   public void reportFighterStatistics () {
   
      // Create the stat bins
      StatBin bin[] = new StatBin[MAX_LEVEL];
      for (int i = 0; i < MAX_LEVEL; i++)
         bin[i] = new StatBin();
   
      // Add all fighters data into stat bins
		for (int i = 0; i < fighterList.size(); i++) {
			Character fighter = (Character) fighterList.get(i);
         int level = fighter.getLevel();
         if (0 <= level && level < MAX_LEVEL) {
            bin[level].addCharacter(fighter);
         }
      }
   
      // Print header for statistics
      System.out.println("Level\tNum\tHP\t"
         + "Str\tInt\tWis\tDex\tCon\tCha");
   
      // Print table of bin statistics
      for (int i = 0; i < MAX_LEVEL; i++) {
         if (bin[i].size() > 0) {
            System.out.print(i + "\t");
            bin[i].printAllStats();
         }
      }
      System.out.println();
   }

   /**
   *  Prints number killed by each monster type.
   */
   void reportMonsterKills () {
      if (REPORT_MONSTER_KILLS) {
         MonsterList ml = MonsterList.getInstance();
         for (int level = 1; level <= 6; level++) {
            System.out.println("Level " + level + " Monsters");
            for (int i = 0; i < ml.size(); i++) {
               Monster mon = ml.get(i);
               if (mon.getTableLevel() == level) {
                  System.out.println(mon.getType() 
                     + ": " + mon.getKillTally());
               }
            }
            System.out.println();
         }
         System.out.println();
      }
   }

   /**
   *  Counts fighters at a given level or above.
   */
	public int countMenAboveLevel (int level) {
		int count = 0;
		for (int i = 0; i < fighterList.size(); i++) {
			if (fighterList.get(i).getLevel() >= level) {
				count++;
			}
		}
		return count;
	}

   /**
   *  Prints fighters at a given level or above.
   */
	public void printMenAboveLevel (int level) {
		fighterList.sortMembers();
		for (int i = fighterList.size() - 1; i >= 0; i--) {
			Monster fighter = fighterList.get(i);
			if (fighter.getLevel() >= level) {
				System.out.println(fighter + "\n");
			}
		}
	}

   /**
   *  Set the base armor type.
   */
	public void setBaseArmor (Armor.Type type) {
		baseArmorType = type;
	}

   /**
   *  Set the base weapon type.
   */
	public void setBaseWeapon (Weapon weapon) {
		baseWeaponType = weapon;	
	}

   /**
   *  Main application method.
   */
   public static void main (String[] args) {
      Arena a = new Arena();
		a.runArena();
		Toolkit.getDefaultToolkit().beep();
   }
}

