import java.util.ArrayList;

/******************************************************************************
*  One character (player or non-player personae).
*
*  @author   Daniel R. Collins (dcollins@superdan.net)
*  @since    2014-05-20
*  @version  1.0
******************************************************************************/

public class Character extends Monster {

	//--------------------------------------------------------------------------
	//  Constants
	//--------------------------------------------------------------------------

	/** Base armor class. */
	static final int BASE_ARMOR_CLASS = 9;

	/** Base movement. */
	static final int BASE_MOVEMENT = 12;

	/** Percent per level for NPC magic items. */
	static final int NPC_PCT_MAGIC_PER_LEVEL = 5;

	/** Whether we apply the prime-requisite XP bonus. */
	static final boolean APPLY_BONUS_XP = false;

	/** Dice for ability scores. */
	static final Dice ABILITY_DICE = new Dice(3, 6);

	//--------------------------------------------------------------------------
	//  Fields
	//--------------------------------------------------------------------------

	/** The six ability scores. */
	int[] abilityScore;

	/** List of classes with XP scores. */
	ArrayList<ClassRecord> classList;

	/** Armor worn. */
	Armor armor;

	/** Shield held. */
	Armor shield;

	/** Weapon in hand. */
	Weapon weaponInHand;	
	
	/** Equipment carried. */
	ArrayList<Equipment> equipList;

	/** Feats acquired. */
	ArrayList<Feat> featList;
	
	/** Allow feats? */
	static boolean allowFeats = false;

	//--------------------------------------------------------------------------
	//  Enumeration
	//--------------------------------------------------------------------------

	/** 
	*  Optional OED fighter feats. 
	*  Caution: Not all are implemented in code at this time.
	*/
	public enum Feat {
		GreatCleave, GreatFortitude, GreatStrength, IronWill, 
		MountedCombat, RapidShot, RapidStrike, Running, Toughness, 
		Tracking, WeaponSpecialization, Raging;
		
		/** Total number of feats available. */
		public static final int length = Feat.values().length;
	};

	//--------------------------------------------------------------------------
	//  Constructors
	//--------------------------------------------------------------------------

	/**
	*  Constructor (single class, no equipment).
	*/
	public Character (String name, String type, 
			String className, int level, Alignment align) {
		super(name, type, align, BASE_ARMOR_CLASS, 
			BASE_MOVEMENT, new Dice(6), null);

		assert(level >= 0); 
		createRandomAbilityScores();
		if (align == null) 
			align = Alignment.random();
		classList = new ArrayList<ClassRecord>(1);
		equipList = new ArrayList<Equipment>(4);
		featList = new ArrayList<Feat>(0);
		ClassType classType = ClassIndex.getInstance()
			.getTypeFromName(className);
		classList.add(new ClassRecord(this, classType, level));
		updateLevelChanges();
		setHitPointsToMax();
	}
	
	//--------------------------------------------------------------------------
	//  Methods
	//--------------------------------------------------------------------------

	// Implement null Monster equipment methods
	public Armor getArmor () { return armor; }
	public Armor getShield () { return shield; }
	public Weapon getWeapon () { return weaponInHand; }

	// Set equipment carried
	public void addEquipment (Equipment equip) { equipList.add(equip); }
	public void dropAllEquipment () { equipList.clear(); }

	/**
	*  Creates random ability scores.
	*/
	void createRandomAbilityScores () {
		abilityScore = new int[Ability.length];
 		for (int i = 0; i < Ability.length; i++) {
 			abilityScore[i] = ABILITY_DICE.roll();
 		}
	}

	/**
	*  Get an ability score bonus/modifier.
	*/
	int getAbilityBonus (Ability ability) {
		int score = getAbilityScore(ability);
		return Ability.getBonus(score);
	}

	/**
	*  Get an ability score.
	*/
	int getAbilityScore (Ability ability) {
		int score = abilityScore[ability.ordinal()];
		if (ability == Ability.Str && hasFeat(Feat.GreatStrength)) {
			score += 2;
		}
		return score;
	}

	/**
	*  Get highest level class.
	*/
	public ClassRecord getTopClass() {
		ClassRecord topClass = classList.get(0);
		for (ClassRecord cr: classList) {
			if (cr.getLevel() > topClass.getLevel())
				topClass = cr;
		}
		return topClass;
	}

	/**
	*  Get highest class level.
	*/
	public int getLevel () {
		return getTopClass().getLevel();
	}

	/**
	*  Get raw number of hit dice.
	*  When "HD" referenced in RPG rules, it means "level" for PC/NPCs.
	*  E.g.: Attacks, saves, magic effects, XP awards.
	*/
	public int getHitDiceNum () {
		return Math.max(getLevel(), 1);
	}

	/**
	*  Update AC, MV on any armor change.
	*/
	void updateArmorStats() {
		armorClass = computeArmorClass();	
		moveInches = computeMoveInches();
	}

	/**
	*  Update attacks on any weapon change.
	*/
	void updateAttackStats() {
		attack = computeAttack();
	}

	/**
	*  Get adjusted armor class.
	*/
	int computeArmorClass() { 
		int AC = BASE_ARMOR_CLASS - getAbilityBonus(Ability.Dex);
		if (armor != null) {
			AC = AC - armor.getBaseArmor() - armor.getMagicBonus();
		}	
		if (shield != null) {
			AC = AC - shield.getBaseArmor() - shield.getMagicBonus();
		}
		return AC;
	}

	/**
	*  Get movement based on armor.
	*/
	int computeMoveInches() {
		return (armor == null ? 
			BASE_MOVEMENT : armor.getMaxMove());
	}

	/**
	*  Get weapon-based melee attack routine.
	*/
	Attack computeAttack () { 
		if (weaponInHand == null) {
			return null;
		}	
		else {
			String name = weaponInHand.getName();
			int rate = (hasFeat(Feat.RapidStrike) ? 2 : 1);
			int atkBonus = baseAttackBonus() + getAbilityBonus(Ability.Str) 
				+ weaponInHand.getMagicBonus();
			Dice damageDice = new Dice(weaponInHand.getBaseDamage());
			int damageAdd = damageDice.getAdd() + getAbilityBonus(Ability.Str) 
				+ weaponInHand.getMagicBonus();
			if (hasFeat(Feat.WeaponSpecialization)) {
				atkBonus += 1;
				damageAdd += 2;
			}
			damageDice.setAdd(damageAdd);
			return new Attack(name, rate, atkBonus, damageDice);
		}	
	}

	/**
	*  Returns base attack bonus (max over all classes).
	*/
	int baseAttackBonus () {
		int maxBAB = 0;
		for (ClassRecord record: classList) {
			if (record.attackBonus() > maxBAB)
				maxBAB = record.attackBonus();
		}
		return maxBAB;
	}

	/**
	*  Find what level of magic-to-hit we can strike.
	*/
	public int getMagicHitLevel () {
		return (weaponInHand == null ? 0 : weaponInHand.getMagicBonus());
	}

	/**
	*  Set armor worn.
	*/
	public void setArmor (Armor armor) { 
		this.armor = armor; 
		updateArmorStats();
	}

	/**
	*  Set shield carried.
	*/
	public void setShield (Armor shield) { 
		this.shield = shield; 
		updateArmorStats();
	}

	/**
	*  Draw a particular weapon from equipment.
	*/
	public void drawWeapon (Weapon weapon) {
		assert(weaponInHand == null);
		if (weapon.getHandsUsed() > 1 && shield != null) {
			equipList.add(shield);
			shield = null;		
		}
		equipList.remove(weapon);
		weaponInHand = weapon;
		updateAttackStats();
	}

	/**
	*  Sheathe weapon in hand back to equipment.
	*/
	public void sheatheWeapon () {
		assert(weaponInHand != null);
		equipList.add(weaponInHand);	
		weaponInHand = null;
		updateAttackStats();		
	}

	/**
	*  Draw best weapon against a given monster.
	*/
	public void drawBestWeapon (Monster monster) {
		if (weaponInHand != null) {
			sheatheWeapon();
		}
		drawWeapon(bestWeaponAgainst(monster));	
	}

	/**
	*  Decide on the best weapon against a given monster.
	*  Considers only weapons in equipment list.
	*/
	Weapon bestWeaponAgainst (Monster monster) {
		assert(weaponInHand == null);
		int maxDamage = -1;
		Weapon bestWeapon = null;
		for (Equipment equip: equipList) {
			if (equip instanceof Weapon) {
				weaponInHand = (Weapon) equip;
				updateAttackStats();
				int damage = maxDamageVsMonster(monster);
				if (damage > maxDamage) {
					maxDamage = damage;
					bestWeapon = weaponInHand;
				}							
			}		
		}	
		weaponInHand = null;
		return bestWeapon;
	}

	/**
	*  Compute max damage against a given monster.
	*/
	int maxDamageVsMonster (Monster monster) {
		Attack atk = getAttack();
		if (atk == null) return 0;
		if (monster != null && !canAttack(monster)) return 0;
		return atk.getDamage().maxRoll();		
	}

	/**
	*  Returns total XP (sum over all classes).
	*/
	public int totalXP () {
		int sum = 0;
		for (ClassRecord record: classList) {
			sum += record.getXP();
		}
		return sum;
	}

	/**
	*  Add XP to the first character class.
	*/
	public void addXP (int xp) {
		addXP(xp, 0);	
	}

	/**
	*  Add XP to a specified character class.
	*/
	public void addXP (int XP, int classIdx) {
		ClassRecord cr = classList.get(classIdx);
		XP = adjustForPrimeReq(XP, cr);
		cr.addXP(XP);
		updateLevelChanges();
	}

	/**
	*  Adjust XP award for the class prime requisite.
	*/
	int adjustForPrimeReq (int XP, ClassRecord cr) {
		if (!APPLY_BONUS_XP) {
			return XP;
		}
		else {
			Ability primeReq = cr.getClassType().getPrimeRequisite();
			int score = getAbilityScore(primeReq);
			int bonusPct = Ability.bonusPercentXP(score);
			return XP + XP * bonusPct/100;
		}
	}

	/**
	*  Lose a level (e.g., energy drain).
	*/
	public void loseLevel () {
		int damage = maxHitPoints - hitPoints;
		getTopClass().loseLevel();
		updateLevelChanges();
		if (SpecialAbility.ENERGY_DRAIN_CONSTANT_DAMAGE) {
			hitPoints = maxHitPoints - damage;
			boundHitPoints();
		}
	}

	/**
	*  Update records after a level change.
	*/
	void updateLevelChanges () {
		hitDice = getTopClass().getHitDice();
		maxHitPoints = findSupMaxHitPoints();
		if (hasFeat(Feat.Toughness)) {
			maxHitPoints += 2 * getLevel();
		}		
		boundHitPoints();
		updateAttackStats();
	}

	/**
	*  Find maximum hit points (supremum over all classes).
	*/
	int findSupMaxHitPoints () {
		int supMaxHP = 0;
		for (ClassRecord record: classList) {
			if (record.getHitPoints() > supMaxHP)
				supMaxHP = record.getHitPoints();		
		}
		return supMaxHP;
	}

	/**
	*  Create human NPC with equipment from class title.
	*/
	static public Character newNPCFromTitle (String title, Alignment align) {
		ClassType classType = ClassIndex.getInstance().getTypeFromTitle(title);
		if (classType != null) {
			int level = classType.getLevelFromTitle(title);
			Character c = new Character(null, "Human", classType.getName(), level, align);
			c.setBasicEquipment();
			for (int i = 1; i <= level; i++) {
				c.checkMagicArmsBoost(NPC_PCT_MAGIC_PER_LEVEL);			
			}
			return c;
		}
		return null;
	}

	/**
	*  Set basic equipment by top class.
	*/
	public void setBasicEquipment () {
		String className = getTopClass().getClassType().getName();
		if (className.equals("Fighter")) {
			if (getLevel() >= 1) {
				setArmor(Armor.makeType(Armor.Type.Plate));
			}
			else {
				setArmor(Armor.makeType(Armor.Type.Leather));
			}
			setShield(Armor.makeType(Armor.Type.Shield));
			addEquipment(Weapon.randomPrimary());
			addEquipment(Weapon.randomSecondary());
		}
		else if (className.equals("Thief")) {
			setArmor(Armor.makeType(Armor.Type.Leather));
			addEquipment(new Weapon("Sword", new Dice(8), 1));
		}
		else {
			addEquipment(Weapon.silverDagger());
		}
	}

	/**
	*  Make rolls to possibly boost magic arms.
	*/
	public void checkMagicArmsBoost (int percent) {
		if (weaponInHand == null) {
			drawBestWeapon(null);
		}
		if (armor != null & Dice.roll(100) <= percent) {
			armor.incMagicBonus();
			updateArmorStats();
		}
		if (shield != null & Dice.roll(100) <= percent) {
			shield.incMagicBonus();			
			updateArmorStats();
		}
		if (weaponInHand != null & Dice.roll(100) <= percent) {
			weaponInHand.incMagicBonus();
			updateAttackStats();
		}
	}

	/**
	*	Roll a saving throw with modifier.
	*/
	public boolean	rollSave	(SavingThrows.SaveType type, int modifier) {
		ClassRecord bestClass = bestClassForSave(type);
		if (type == SavingThrows.SaveType.Death && 
				hasFeat(Feat.GreatFortitude))
			modifier += 4;		
		return SavingThrows.getInstance().rollSave(
			type, bestClass.getClassType().getSaveAsClass(), 
			bestClass.getLevel(), modifier); 
	}

	/**
	*	Find the best class to use for a given saving throw.
	*/
	ClassRecord bestClassForSave (SavingThrows.SaveType saveType) {
		ClassRecord bestClass = null;
		int bestScore = Integer.MAX_VALUE;
		for (ClassRecord record: classList) {
			String saveAs = record.getClassType().getSaveAsClass();
			int score = SavingThrows.getInstance().getSaveTarget(
				saveType, saveAs, record.getLevel());
			if (score < bestScore) {
				bestScore = score;
				bestClass = record;			
			}
		}
		return bestClass;
	}	

	/**
	*  Set if we should be using optional feats.
	*/
	static public void setFeatUsage (boolean permit) {
		allowFeats = permit;	
	}

	/**
	*  Are we allowing optional feats?
	*/
	static public boolean allowFeats () {
		return allowFeats;	
	}
	
	/**
	*  Does this character have a given feat?
	*/
	public boolean hasFeat (Feat feat) {
		return featList.contains(feat);
	}

	/**
	*  Add a random feat to this character.
	*/
	public void addFeat () {
		if (allowFeats) {
			while	(true) {
				int rand = Dice.roll(Feat.length) - 1;
				Feat newFeat = Feat.values()[rand];
				if (!hasFeat(newFeat)) {
					featList.add(newFeat);
					return;
				}
			}	
		}
	}

	/**
	*  Lose the last feat for this character.
	*/
	public void loseFeat () {
		if (featList.size() > 1) {
			featList.remove(featList.size() - 1);
		}	
	}

	/**
	*  Identify this object as a string.
	*/
	public String toString () {

		// Header
		String s = (name == null ? "" : name + ", ");
		s += (type == null ? "" : type + " ");
		s += classString() + " ";

		// Stat block
		s += "(AC " + getAC() + ", MV " + getMV() + ", HD " + getHD()
			+ ", hp " + getHP() + ", Atk " + getAttack() + "; ";
		s += abilityString();

		// Feats
		String featStr = featString();
		if (featStr.length() > 0) {
			s += "; " + featStr;		
		}

		// Equipment
		String equipStr = equipString();
		if (equipStr.length() > 0) {
			s += "; " + equipStr;		
		}
		s += ")";
		return s;		
	}	

	/**
	*  String representation of all class and levels.
	*/
	String classString () {
		String s = "";
		for (ClassRecord record: classList) {
			s += record.getClassType().getAbbreviation() 
				+ record.getLevel() + "/";
		}
		return s.substring(0, s.length() - 1);
	}

	/**
	*  String representation of all ability scores.
	*/
	String abilityString () {
		String s = "";
		for (Ability a: Ability.values()) {
			s += a.name() + " " + getAbilityScore(a) + ", ";
		}
		return s.substring(0, s.length() - 2);
	}

	/**
	*  String representation of feats.
	*/
	String featString () {
		String s = "";
		for (Feat feat: featList) {
			s += formatFeat(feat) + ", ";
		}	
		return (s.length() < 2 ? "" : 	
			s.substring(0, s.length() - 2));
	}

	/**
	*  Format feat name with spaces.
	*/
	String formatFeat (Feat feat) {
		String r = "";
		String s = feat.toString();
		for (int i = 0; i < s.length(); i++) {
			if (i > 0 && 
					java.lang.Character.isUpperCase(s.charAt(i))) {
				r += " ";
			}
			r += s.charAt(i);
		}			
		return r;	
	}

	/**
	*  String representation of equipment.
	*/
	String equipString () {
		String s = "";
		if (armor != null) s += armor + ", ";
		if (shield != null) s += shield + ", ";	
		if (weaponInHand != null) s += weaponInHand + ", ";
		for (Equipment equip: equipList) {
			s += equip + ", ";		
		}
		return (s.length() < 2 ? "" : 	
			s.substring(0, s.length() - 2));
	}

	/**
	*  Main test method.
	*/
	public static void main (String[] args) {
		Dice.initialize();
		Character p = new Character(null, "Human", "Fighter", 1, null);
		p.setArmor(Armor.makeType(Armor.Type.Plate));
		p.setShield(Armor.makeType(Armor.Type.Shield));
		p.addEquipment(Weapon.randomPrimary());
		p.addEquipment(Weapon.randomSecondary());
		p.drawBestWeapon(null);
		System.out.println(p);
	}
	
	/*--------------------------------------------------------------------------
	*  TODO:
	*  Languages, class abilities, other equipment, encumbrance.
	*  Possibly use base Monster.hitDice, as classes overwrite top HD.
	*-------------------------------------------------------------------------*/
}

