/******************************************************************************
*  Record of one class gained by a character (XP, level, hit points, etc.).
*
*  @author   Daniel R. Collins (dcollins@superdan.net)
*  @since    2014-05-22
*  @version  1.0
******************************************************************************/

public class ClassRecord {

	//--------------------------------------------------------------------------
	//  Fields
	//--------------------------------------------------------------------------

	/** Character reference. */
	Character character;

	/** Class type for this class. */
	ClassType classType;

	/** Level in this class. */
	int level;

	/** Hit points earned for this class. */
	int hitPoints;

	/** Experience points earned towards this class. */
	int XP;

	//--------------------------------------------------------------------------
	//  Constructors
	//--------------------------------------------------------------------------

	/**
	*  Constructor (by ClassInfo, level, random hit points).
	*/
	public ClassRecord (Character character, ClassType type, int level) {
		this.character = character;
		this.classType = type;
		this.level = level;
		this.XP = type.getExperienceReq(level);
		rollFullHitPoints();
		addAllFeats();
	}

	//--------------------------------------------------------------------------
	//  Methods
	//--------------------------------------------------------------------------
	public ClassType getClassType () { return classType; }
	public int getLevel () { return level; }
	public int getHitPoints () { return hitPoints; }
	public int getXP () { return XP; }

	/**
	*  Return the base attack bonus for this class/level.
	*/
	public int attackBonus () {
		return classType.getAttackBonus(level);
	}

	/**
	*  Add XP to this class.
	*/
	public void addXP (int awardXP) {
		XP += awardXP;
		if (XP >= classType.getExperienceReqNext(level)) {
			addLevel();
		}
	}

	/**
	*  Increment the level.
	*/
	public void addLevel () {
		level++;	
		boundXPToLevel();
		rollNewHitPoints(level);
		if (isFeatLevel(level)) {
			character.addFeat();
		}
	}

	/**
	*  Decrement the level.
	*/
	public void loseLevel () {

		// At level 0, death/lose class
		if (level < 1) {
			level = -1;
			hitPoints = 0;		
			XP = Integer.MIN_VALUE;
		}

		// Else reduce hit points proportionally
		else {
			if (isFeatLevel(level)) {
				character.loseFeat();
			}
			Dice diceInc = classType.getHitDiceInc(level);
			if (diceInc.getAdd() > 0) {
				hitPoints -= diceInc.getAdd();
			}
			else {
				hitPoints = hitPoints * (level - 1)/level;
			}
			level--;
			boundXPToLevel();
		}
	}

	/**
	*  Choose all feats for new character.
	*/
	void addAllFeats () {
		for (int i = 1; i <= level; i++) {
			if (isFeatLevel(i)) {
				character.addFeat();			
			}
		}
	}

	/**
	*  Is this a level where we gain a feat?
	*/
	boolean isFeatLevel (int newLevel) {
		return Character.allowFeats()
			&& classType.getName().equals("Fighter")
			&& newLevel % 4 == 0;	
	}

	/**
	*  Bound XP to indicated level.
	*/
	void boundXPToLevel () {
		int min = classType.getExperienceReq(level);
		int max = classType.getExperienceReqNext(level) - 1;
		if (XP < min) XP = min;
		if (XP > max) XP = max;	
	}

	/**
	*  Add hit points for a given level.
	*/
	void rollNewHitPoints (int newLevel) {
		int hpBonus = character.getAbilityBonus(Ability.Con);

		// From level 0 to 1st
		if (newLevel == 1) {
			hitPoints += (classType.getHitDiceType() - 6)/2;
			hitPoints = Math.max(hitPoints, 1);
		}
		
		// Levels above 1st
		else if (newLevel > 1) {
			Dice newDice = classType.getHitDiceInc(newLevel);
			hitPoints += (newDice.getNum() > 0 ?
				newDice.rollPlus(hpBonus) : newDice.getAdd());
		}		
	}

	/**
	*  Roll full hit points from start to current level.
	*/
	void rollFullHitPoints () {
		int hpBonus = character.getAbilityBonus(Ability.Con);
		hitPoints = new Dice(1, 6).rollPlus(hpBonus); // Level 0
		for (int newLevel = 1; newLevel <= level; newLevel++) {
			rollNewHitPoints(newLevel);
		}
	}

	/**
	*  Get adjusted hit dice for this class.
	*/
	public Dice getHitDice () {
		Dice dice = new Dice(classType.getHitDiceTotal(level));
		int conBonus = character.getAbilityBonus(Ability.Con);
		int hpBonus = conBonus * dice.getNum();
		dice.setAdd(dice.getAdd() + hpBonus);
		return dice;
	}

	/**
	*  Identify this object as a string.
	*/
	public String toString() {
		return classType.getName() 
		 + "(" + level + ", " + hitPoints + ", " + XP + ")";
	}	
}

