import java.util.ArrayList;

/******************************************************************************
*	Monster (hostile or benign	creature).
*
*	@author	 Daniel R. Collins (dcollins@superdan.net)
*	@since	 2014-05-20
*	@version	 1.0
******************************************************************************/

public class Monster	implements Comparable<Monster> {

	//--------------------------------------------------------------------------
	//	 Constants
	//--------------------------------------------------------------------------

	/** Maximum enemies who can melee us at once. */
	static final int MAX_MELEERS = 6;

	//--------------------------------------------------------------------------
	//	 Fields
	//--------------------------------------------------------------------------

	String name;
	String type;
	Dice numberAppearing;
	int armorClass;
	int moveInches;
	Dice hitDice;
	int inLairPct;
	String treasureType;
	Attack attack;
	Alignment alignment;
	int xpSpecialAwards;
	int tableLevel;
	int hitPoints;
	int maxHitPoints;
	int breathCharges;
	int killTally;
	int timesMeleed;
	Monster host;
	Monster charmer;
	ArrayList<SpecialAbility> specialList;
	ArrayList<SpecialAbility> conditionList;
	
	//--------------------------------------------------------------------------
	//	 Constructors
	//--------------------------------------------------------------------------

	/**
	*	Constructor	(primitive stats).
	*/
	public Monster	(int AC,	int MV, int	HD) {
		this(null, "Monster", Alignment.Neutral, AC,	MV,
			new Dice(HD, 6), new	Attack(HD, 1));
	}

	/**
	*	Constructor	(primitive stats,	attack rate, damage dice).
	*/
	public Monster	(int AC,	int MV, int	HD, int atkRate, int	damDice)	{
		this(null, "Monster", Alignment.Neutral, AC,	MV,
			new Dice(HD, 6), new	Attack(atkRate, HD, damDice));
	}

	/**
	*	Constructor	(most	fields; no special abilities).
	*/
	public Monster	(String name, String	type,	Alignment alignment,	
			int AC, int	MV, Dice	hitDice,	Attack attack)	{
		specialList	= new	ArrayList<SpecialAbility>(); 
		conditionList = new ArrayList<SpecialAbility>(); 
		this.name =	name;
		this.type =	type;
		this.alignment	= alignment;
		this.armorClass =	AC;
		this.moveInches =	MV;
		this.hitDice =	hitDice;
		this.attack	= attack;
		rollHitPoints();
	}

	/**
	*	Constructor	(from	String array).
	*	@param s	The String array.
	*/
	public Monster	(String[] s) {

		//	Special abilities	parsed first
		specialList	= new	ArrayList<SpecialAbility>(3); 
		String parts[]	= s[9].split(", ");
		for (String	part: parts)	{
			if	(part.length()	> 1) {
				SpecialAbility ability = 
					SpecialAbility.createFromString(part);
				if (ability != null) {
					specialList.add(ability);
				}
			}					
		}
		
		//	Other	fields		
		name = null;
		type = s[0];
		numberAppearing =	new Dice(s[1]);
		armorClass = CSVReader.parseInt(s[2]);
		moveInches = CSVReader.parseInt(s[3]);
		hitDice = new Dice(s[4]);
		inLairPct =	CSVReader.parseInt(s[5]);
		treasureType =	s[6];
		int attackRate	= CSVReader.parseInt(s[7]);
		int attackBonus =	hitDice.getNum();
		Dice damageDice =	new Dice(s[8]);
		attack =	new Attack(null, attackRate, 
			attackBonus, damageDice);
		alignment =	Alignment.valueOf(s[10]);
		xpSpecialAwards = countAsterisks(s[11]);		
		tableLevel = CSVReader.parseInt(s[12]);
		conditionList = new ArrayList<SpecialAbility>(); 
		if	(hasSpecial(SpecialAbility.SpecialType.BreathWeapon))	{
			breathCharges = 3;
		}
		rollHitPoints();
	}

	/**
	*	Constructor	(copy).
	*/
	public Monster	(Monster	src) {
		specialList	= new	ArrayList<SpecialAbility>(src.specialList); 
		conditionList = new ArrayList<SpecialAbility>(src.conditionList);	
		name = src.name;
		type = src.type;
		numberAppearing =	src.numberAppearing;
		armorClass = src.armorClass;
		moveInches = src.moveInches;
		hitDice = src.hitDice;
		inLairPct =	src.inLairPct;
		treasureType =	src.treasureType;
		attack =	src.attack;
		alignment =	src.alignment;
		xpSpecialAwards = src.xpSpecialAwards;
		tableLevel = src.tableLevel;
		killTally = src.killTally;
		host = src.host;
		charmer = src.charmer;
		breathCharges = src.breathCharges;
		hitPoints =	src.hitPoints;
		maxHitPoints =	src.maxHitPoints;
	}

	//--------------------------------------------------------------------------
	//	 Methods
	//--------------------------------------------------------------------------

	//	Primary accessors
	public String getName ()	{ return	name;	}
	public String getType ()	{ return	type;	}
	public Dice	getNumberAppearing ()	{ return	numberAppearing; }
	public int getArmorClass ()	{ return	armorClass;	}
	public int getMoveInches ()	{ return	moveInches;	}
	public int getHitPoints () { return hitPoints; }
	public int getMaxHitPoints () { return maxHitPoints; }
	public int getInLairPct () { return inLairPct; }
	public Dice	getHitDice () {	return hitDice; }
	public int getHitDiceNum ()	{ return	hitDice.getNum();	}
	public int getLevel () { return getHitDiceNum(); }
	public String getTreasureType () { return treasureType; }
	public Attack getAttack () { return attack; }
	public Alignment getAlignment () { return alignment; }
	public int getXpSpecialAwards ()	{ return	xpSpecialAwards;	}
	public int getTableLevel ()	{ return	tableLevel;	}
	public int getKillTally ()	{ return	killTally;	}
	public int getTimesMeleed () { return timesMeleed; }
	public Monster	getHost () {	return host; }
	public boolean	isCharmed ()	{ return	charmer != null; }

	//	Shortcut	accessors
	public int getAC () {	return getArmorClass();	}
	public int getHD () {	return getHitDiceNum();	}
	public int getHP () {	return getHitPoints(); }
	public int getMV () {	return getMoveInches();	}

	//	Basic	mutators
	public void	setType (String type) {	this.type =	type;	}
	public void	setAlignment (Alignment	align) {	alignment =	align; }
	public void	clearTimesMeleed ()	{ timesMeleed = 0; }
	public void	incTimesMeleed () { timesMeleed++; }

	//	Null methods for Character	inheritance
	public Weapon getWeapon	()	{ return	null;	}
	public void	drawBestWeapon	(Monster	m)	{}
	public void	sheatheWeapon () {}
	public void checkMagicArmsBoost (int pct) {}
	public void addXP (int xp) {}
	public boolean hasFeat (Character.Feat feat) { return false; }

	/**
	*	Spawn	a new	monster of this type, with	different hit points.
	*/
	public Monster	spawn	()	{
		if	(hasSpecial(SpecialAbility.SpecialType.NPC))	{
			Character c	= Character.newNPCFromTitle(
				getType(), Alignment.Chaotic);
			c.setType("Evil "	+ getType());
			return c;
		}
		else {
			Monster m =	new Monster(this);
			m.rollHitPoints();
			return m;
		}
	}

	/**
	*	Roll hit	points from	hit dice.
	*/
	private void rollHitPoints	()	{
		if	(type.contains("Dragon")) {
			int age = Dice.roll(hitDice.getSides());
			maxHitPoints =	age *	hitDice.getNum();
		}
		else if (hasSpecial(SpecialAbility.SpecialType.Multiheaded)) {
			maxHitPoints =	hitDice.maxRoll();
		}
		else {
			maxHitPoints =	hitDice.roll();
		}
		hitPoints =	maxHitPoints;
	}

	/**
	*	Take damage	(minimum	0 hp).
	*/
	public void	takeDamage (int damage)	{
		hitPoints -= damage;
		boundHitPoints();
		headCount();
	}

	/**
	*	Kill this monster	(reduce to 0 hp).
	*/
	public void	instaKill () {
		hitPoints =	0;
	}

	/**
	*	Check	if	the monster	is	out of the fight.
	*/
	public boolean	horsDeCombat()	{
		return hitPoints <= 0
			||	hasDisablingCondition()
			||	isCharmed();
	}

	/**
	*	Set hit points	to	maxHitPoints.
	*/
	public void	setHitPointsToMax	()	{
		hitPoints =	getMaxHitPoints();
	}

	/**
	*	Bound	current hit	points.
	*/
	public void	boundHitPoints	()	{
		int min = 0;
		int max = getMaxHitPoints();
		if	(hitPoints < min)	hitPoints =	min;
		if	(hitPoints > max)	hitPoints =	max;
	}

	/**
	*	Check if we are subject to more melee attacks.
	*/
	public boolean isOpenToMelee () { 
		return !horsDeCombat() && (timesMeleed < MAX_MELEERS); 
	}

	/**
	*	Carry	out full	attack routine	on	another creature.
	*/
	public void	fullAttack (Attack attack,	Monster target) {
		for (int	i = 0; i	< attack.getRate(); i++) {
			boolean last =	(i	==	attack.getRate() - 1); 
			singleAttack(attack,	target, last);
		}
	}

	/**
	*	Make one	attack on another	creature.
	*/
	public void	singleAttack (Attack	attack, Monster target,	boolean last) {
		int naturalRoll =	Dice.roll(20);
		int totalRoll = naturalRoll +	attack.getBonus()	
			+ target.getAC() + hitModifier(target);
		if	(canAttack(target) && (naturalRoll == 20 || totalRoll	>=	20)) {
			int damage = attack.getDamage().roll();
			damage =	checkDamageReduction(target, damage);
			target.takeDamage(damage);
			checkSpecialOnHit(target, totalRoll, last);
		}
	}

	/**
	*	Can we feasibly attack this target?
	*/
	public boolean	canAttack (Monster target)	{
		SpecialAbility	special;

		//	Check	silver to hit
		special = target.findSpecial(SpecialAbility.SpecialType.SilverToHit);
		if	(special	!=	null)	{
			boolean silverWeapon	= (getWeapon()	!=	null 
				&&	getWeapon().getMaterial() == Weapon.Material.Silver);
			return silverWeapon || getMagicHitLevel()	> 0;
		}

		//	Check	magic	to	hit
		special = target.findSpecial(SpecialAbility.SpecialType.MagicToHit);
		if	(special	!=	null)	{
			return getMagicHitLevel() >= special.getParam();
		}
		return true;
	}

	/**
	*	Find what level of magic-to-hit we can	strike.
	*/
	public int getMagicHitLevel () {

		//	Check	if	we	are also	magic-to-hit
		SpecialAbility	special = findSpecial(SpecialAbility.SpecialType.MagicToHit);
		int magicHitLevel	= (special == null ?	0 : special.getParam());

		//	Check	if	we	have high hit dice
		int fractionHD	= getHD() /	4;
		return Math.max(magicHitLevel, fractionHD);
	}

	/**
	*	Return hit modifier against a	target.
	*	@return Modifier to hit.
	*/
	public int hitModifier (Monster target) {
		int modifier =	0;

		//	Berserker bonus
		if	(hasSpecial(SpecialAbility.SpecialType.Berserk) ||
				hasFeat(Character.Feat.Raging)) {
			modifier	+=	2;
		}
	
		// Gaze attack from target; asssume we look aside
		if	(target.hasSpecial(SpecialAbility.SpecialType.PetrifyGaze))	{
			modifier	-=	4;
		}
		
		return modifier;
	}

	/**
	*	Check	for damage reduction	on	target.
	*	@return Adjusted damage.
	*/
	public int checkDamageReduction (Monster target, int damage) {
		if	(target.hasSpecial(SpecialAbility.SpecialType.DamageReduction)) {
			damage /= 2;
		}
		return damage;
	}

	/**
	*	Check	for special	ability triggers when we hit.
	*/
	public void	checkSpecialOnHit	(Monster	target, int	totalRoll, boolean isLastAttack)	{
		for (SpecialAbility s: specialList)	{
			switch (s.getType()) {

				case Poison:			
					if	(isLastAttack &&
							!target.rollSave(SavingThrows.SaveType.Death, s.getParam())) {
						target.addCondition(SpecialAbility.SpecialType.Poison);
					}
					break;
			
				case Paralysis:			
					if	(!target.rollSave(SavingThrows.SaveType.Paralysis)) {
						target.addCondition(SpecialAbility.SpecialType.Paralysis);
					}
					break;
			
				case PetrifyTouch: 			
					if	(!target.rollSave(SavingThrows.SaveType.Petrify)) {
						target.addCondition(SpecialAbility.SpecialType.Petrification);
					}
					break;

				case BloodDrain: 
				case Constrict:
					setHost(target);
					break;
					
				case EnergyDrain:
					if	(!s.allowSave() || 
							!target.rollSave(SavingThrows.SaveType.Death)) {
						for (int	j = 0; j	< s.getParam(); j++)	{
							target.loseLevel();
						}
					}
					break;

				case Rotting: 
					if	(!s.allowSave() || 
							!target.rollSave(SavingThrows.SaveType.Death)) {
						target.addCondition(SpecialAbility.SpecialType.Rotting);
					}
					break;

				case Immolate: 
					if (isLastAttack && new Dice(2, 6).roll() >= 7) {
						int damage = new Dice(3, 6).roll();
						boolean saved = target.rollSave(SavingThrows.SaveType.Breath);
						target.takeDamage(saved	? damage/2 : damage);
					}
					break;

				case Swallow:
					if (!isLastAttack && totalRoll > 24) {
						target.addCondition(SpecialAbility.SpecialType.Swallow);
					}
					break;
			}
		}
	}

	/**
	*	Make a special	attack on an enemy party.
	*	Ranged/outside	of	melee	specials	should go here.
	*  Assumes only one of these abilities are possessed.
	*/
	void makeSpecialAttack (Party enemy) {
		for (SpecialAbility s: specialList)	{
			switch (s.getType()) {
			
				case HurlRocks: 
					Monster targetRock = enemy.random();
					Attack attackRock = new Attack("Hurl Rock", 1, getHD(), new Dice(2, 6));
					singleAttack(attackRock, targetRock, false);
					break;

				case TailSpikes:
					Attack attackSpike = new Attack("Tail Spike", 1, getHD(), new Dice(1, 6));
					for (int i = 0; i < 6; i++) {
						Monster targetSpike = enemy.random();
						singleAttack(attackSpike, targetSpike, false);
					}
					break;

				case Charm:
					Monster targetCharm = enemy.random();
					int modifier = s.getParam();
					if (hasFeat(Character.Feat.IronWill)) modifier += 4;
					if (!targetCharm.rollSave(SavingThrows.SaveType.Petrify, modifier)) {
						targetCharm.setCharmed(this);
					}
					break;

				case PetrifyGaze:
					for (int i = 0; i < enemy.size(); i++) {
						Monster targetGaze = enemy.get(i);
						if	(!targetGaze.rollSave(SavingThrows.SaveType.Petrify)) {
							targetGaze.addCondition(SpecialAbility.SpecialType.Petrification);
						}
					}
					break;

				case Fear:
					for (int i = 0; i < enemy.size(); i++) {
						Dice dice = new Dice(2, 6); 
						Monster targetFear = enemy.get(i);
						if	(dice.rollPlus(targetFear.getHD()) < 6) {
							targetFear.addCondition(SpecialAbility.SpecialType.Fear);
						}
					}
					break;
			}					
		}
	}

	/**
	*	Find if we have a	given	type of special ability.
	*/
	public SpecialAbility findSpecial (SpecialAbility.SpecialType type) {
		for (SpecialAbility s: specialList)	{
			if	(s.getType() == type) {
				return s;
			}
		}
		return null;
	}

	/**
	*	Check	if	this monster has a given type	of	special ability.
	*/
	public boolean	hasSpecial (SpecialAbility.SpecialType	type)	{
		return findSpecial(type) != null;
	}

	/**
	*	Add a	condition suffered from	a special ability.
	*/
	public void	addCondition (SpecialAbility.SpecialType type) {
		conditionList.add(new SpecialAbility(type));
	}

	/**
	*	Find if we suffer	from a given condition.
	*/
	public SpecialAbility findCondition	(SpecialAbility.SpecialType type) {
		for (SpecialAbility s: conditionList) {
			if	(s.getType() == type) {
				return s;
			}
		}
		return null;
	}

	/**
	*	Check	if	we suffer from a given condition.
	*/
	public boolean	hasCondition (SpecialAbility.SpecialType type) {
		return findCondition(type)	!=	null;	
	}

	/**
	*	Check	if	we suffer from a disabling condition.
	*/
	public boolean	hasDisablingCondition () {
		for (SpecialAbility s: conditionList) {
			if	(s.isDisabling())	{
				return true;
			}
		}
		return false;
	}

	/**
	*	Regenerate hit	points if appropriate.
	*/
	public void	checkRegeneration	()	{
		SpecialAbility	special = findSpecial(SpecialAbility.SpecialType.Regeneration);
		if	(special	!=	null)	{
			hitPoints += special.getParam();
			boundHitPoints();
		}
	}

	/**
	*  Drain blood from host if appropriate.
	*	@return Did	we	drain	blood?
	*/
	public boolean	checkDrainBlood () {
		SpecialAbility	special = findSpecial(SpecialAbility.SpecialType.BloodDrain);
		if	(special	!=	null && host != null) {
			int drain =	new Dice(1,	6).roll();
			host.takeDamage(drain);
			if	(host.horsDeCombat()) {
				host = null;
			}
			return true;
		}
		return false;
	}

	/**
	*	Constrict host	if	appropriate.
	*	@return Did	we	constrict?
	*/
	public boolean	checkConstriction	()	{
		SpecialAbility	special = findSpecial(SpecialAbility.SpecialType.Constrict);
		if	(special	!=	null && host != null) {
			int damage = new Dice(1, 6).roll();
			host.takeDamage(damage);
			if	(host.horsDeCombat()) {
				host = null;			
			}
			return true;
		}
		return false;
	}

	/**
	*	Check	for a	breath weapon attack.
	*  Estimates max number hit for given area.
	*	@return Did	we	make a breath attack?
	*/
	public boolean	checkBreathWeapon	(Party enemy) {
		SpecialAbility	breath =	findSpecial(SpecialAbility.SpecialType.BreathWeapon);
		if	(breath != null && breathCharges	> 0 && new Dice(2, 6).roll() >= 7) {

			//	Parameter 1: Red dragon, 2: Chimera, 3: Gorgon
			switch (breath.getParam())	{
				case 1: breatheFireOnEnemy(enemy, 14, maxHitPoints); break;
				case 2: breatheFireOnEnemy(enemy, 4, new Dice(3, 6).roll()); break;
				case 3: breatheStoneOnEnemy(enemy, 7); break;
			}
			breathCharges--;
			return true;
		}
		return false;
	}

	/**
	*	Apply area flame-breath damage to an enemy party.
	*  Assumes we hit about half, up to max number.
	*/
	void breatheFireOnEnemy (Party enemy, int maxNum, int damage) {
		int number = Math.min((enemy.size()+1)/2, maxNum); 
		ArrayList<Monster> group = enemy.randomGroup(number);
		for (Monster m: group) {
			boolean saved = m.rollSave(SavingThrows.SaveType.Breath);
			m.takeDamage(saved ? damage/2 : damage);
		}
	}

	/**
	*	Apply area petrification effect on an enemy party.
	*  Assumes we hit about half, up to max number.
	*/
	void breatheStoneOnEnemy (Party enemy, int maxNum) {
		int number = Math.min((enemy.size()+1)/2, maxNum); 
		ArrayList<Monster> group = enemy.randomGroup(number);
		for (Monster m: group) {
			if (!m.rollSave(SavingThrows.SaveType.Petrify)) {
				m.addCondition(SpecialAbility.SpecialType.Petrification);			
			}
		}
	}

	/**
	*	Check	if we can summon minions to a party.
	*/
	public void checkSummoning (Party party) {
		SpecialAbility summons = findSpecial(SpecialAbility.SpecialType.Summoning);
		if (summons != null) {
			MonsterList	list = MonsterList.getInstance();
			Monster monster =	list.getByType("Wolf");
			if	(monster	!=	null)	{
				int number = new Dice(3, 6).roll();
				for (int i = 0; i < number; i++) {
					party.add(monster.spawn());
				}
			}
		}
	}

	/**
	*	Attach ourselves to some creature (e.g., blood drain).
	*/
	public void	setHost (Monster host) {
		this.host =	host;	
	}

	/**
	*	Indicate	that we are	charmed by some other creature.
	*/
	public void	setCharmed (Monster charmer) {
		this.charmer =	charmer;
	}

	/**
	*	Count	current heads for	multiheaded	types.
	*	Set one attack	per full	hit die.
	*/
	void headCount	()	{
		if	(hasSpecial(SpecialAbility.SpecialType.Multiheaded)) {
			int hitDieSides =	getHitDice().getSides();
			int newRate	= (getHP() - 1) /	hitDieSides	+ 1;
			getAttack().setRate(newRate);
		}
	}

	/**
	*	Lose a level (e.g., energy	drain).
	*/
	public void	loseLevel () {
		int HD =	getHitDiceNum();
		if	(HD <= 1) {
			maxHitPoints =	0;
			hitDice.setNum(0);
		}
		else {
			maxHitPoints =	maxHitPoints *	(HD - 1)/HD;
			hitDice.setNum(HD	- 1);
		}
		boundHitPoints();
	}

	/**
	*	Roll a saving throw with no modifier.
	*/
	public boolean	rollSave	(SavingThrows.SaveType type) {
		return rollSave(type, 0);
	}

	/**
	*	Roll a saving throw with modifier.
	*/
	public boolean	rollSave	(SavingThrows.SaveType type, int modifier) {
		return SavingThrows.getInstance().rollSave(
			type, "Fighter", getHD(), modifier);
	}

	/**
	*	Add to kill tally.
	*/
	public void addToKillTally (int num) {
		killTally += num;	
	}

	/**
	*	Compare to another monster	by	hit dice.
	*/
	public int compareTo	(Monster	other) {
		return Integer.compare(getHD(), other.getHD());
	 }

	/**
	*	Count asterisks in a string (e.g., for XP bonus).
	*/
	int countAsterisks (String s) {
		int count = 0;
		for (int i = 0; i < s.length(); i++) {
			if (s.charAt(i) == '*') count++;		
		}	
		return count;
	}

	/**
	*	Get the XP award value for defeating this monster.
	*/
	int getXPAward () {
		return XPTable.getInstance().getXPAward(this);
	}

	/**
	*	Identify	this object	as	a string.
	*/
	public String toString() {
		return (getName()	!=	null ? getName() + ", "	: "")
		  + getType() + " (AC "	+ getAC() +	", MV " + getMV()	
		  + ", HD "	+ getHD() +	", hp " + getHP()	
		  + ", Atk " +	getAttack().getRate() 
		  + ", Dam " +	getAttack().getDamage()
		  + (specialList.size()	> 0 ?	"; " + specialString() : "")
		  + ")";
	}

	/**
	*	Identify	special abilities	as	a string.
	*/
	public String specialString()	{
		String s	= "";
		for (SpecialAbility special: specialList) {
			s += special +	", ";
		}
		return s.substring(0, s.length()	- 2);
	}

	/**
	*	Main test method.
	*/
	public static void main	(String[] args) {
		Monster c =	new Monster(5,	9,	4);
		System.out.println(c);
	}

	/*--------------------------------------------------------------------------
	*	TODO:
	*  Distinct attack forms (in ArrayList, characters might register/swap)
	*-------------------------------------------------------------------------*/
}

