/******************************************************************************
*  Special abilities for monsters.
*
*  @author   Daniel R. Collins (dcollins@superdan.net)
*  @since    2016-01-20
*  @version  1.0
******************************************************************************/

public class SpecialAbility {

	//--------------------------------------------------------------------------
	//  Constants
	//--------------------------------------------------------------------------

	/** Allow a save versus energy drain? */
	static final boolean ENERGY_DRAIN_SAVE = false;

	/** Allow a save versus rotting disease? */
	static final boolean ROTTING_DISEASE_SAVE = false;

	/** Keep constant damage on energy drain (more punitive?) */
	public static final boolean ENERGY_DRAIN_CONSTANT_DAMAGE = true;

	//--------------------------------------------------------------------------
	//  Enumeration
	//--------------------------------------------------------------------------

	/**	
	*  Enumeration for different special ability types.
	*  Caution: Not all are implemented in code at this time.
	*/
	public enum SpecialType {
		Poison, Paralysis, Berserk, BloodDrain, EnergyDrain,
		Fly, SilverToHit, MagicToHit, Regeneration, 
		DamageReduction, Rotting, TailSpikes, Multiheaded, 
		PetrifyTouch, PetrifyGaze, HurlRocks, 
		Fear, BreathWeapon, SeeInvisible, Immolate, 
		Swallow, Spells, MagicResistance, Summoning, 
		Polymorph, Charm, Petrification, Lycanthropy,
		Constrict, NPC
	}

	//--------------------------------------------------------------------------
	//  Fields
	//--------------------------------------------------------------------------

	/** Type of special ability. */
	private SpecialType type;

	/** Parameter for level of ability. */
	private int param;

	//--------------------------------------------------------------------------
	//  Constructors
	//--------------------------------------------------------------------------

	/**	
	*  Constructor (no param).
	*/
	SpecialAbility (SpecialType type) {
		this(type, 0);
	}

	/**	
	*  Constructor (all fields).
	*/
	SpecialAbility (SpecialType type, int param) {
		this.type = type;
		this.param = param;
	}
	
	//--------------------------------------------------------------------------
	//  Methods
	//--------------------------------------------------------------------------
	public SpecialType getType() { return type; }
	public int getParam() { return param; }

	/**
	*  Create new special ability from string.
	*/
	static public SpecialAbility createFromString (String s) {

		// Search for type
		SpecialType type = null;
		for (SpecialType t: SpecialType.values()) {
			if (s.startsWith(t.name())) {
				type = t;
			}
		}

		// Check if type not found
		if (type == null) {
			System.out.println("Warning, special ability not found: " + s);
			return null;		
		}
		
		// Look inside parentheses for param
		int param = 0; 
		int start = s.indexOf('(');
		int end = s.indexOf(')');
		if (start >= 0 && end > start) {
			String contents = s.substring(start + 1, end);
			param = Integer.parseInt(contents);
		}
		
		// Return this special ability
		return new SpecialAbility(type, param);
	}	

	/**
	*  Do we optionally allow a save against this ability?
	*/
	public boolean allowSave () {
		switch (type) {
			case EnergyDrain: return ENERGY_DRAIN_SAVE;
			case Rotting: return ROTTING_DISEASE_SAVE;
			default: return true;
		}	
	}
	
	/**
	*  Does this ability confer a disabling condition?
	*/
	public boolean isDisabling () {
		return type == SpecialType.Poison 
			|| type == SpecialType.Paralysis
			|| type == SpecialType.Petrification 
			|| type == SpecialType.Swallow
			|| type == SpecialType.Fear;
	}

	/**
	*  Identify this object as a string.
	*/
	public String toString() {
		String s = type.name();
		if (param != 0) {
			s += " (" + param + ")";		
		}
		return s;
	}
	
	/**
	*  Main test function.
	*/
	public static void main (String[] args) {
		System.out.println(createFromString("Poison"));
		System.out.println(createFromString("Poison (2)"));
		System.out.println(createFromString("Poison (4)"));
		System.out.println(createFromString("EnergyDrain (1)"));
		System.out.println(createFromString("EnergyDrain (2)"));
		System.out.println(createFromString("Regeneration (3)"));
	}
}

